// vim:ts=4:sw=4:

#include <awds/TcpServer.h>

#include <sys/socket.h>
#include <netdb.h>
#include <string.h>
#include <errno.h>
#include <sstream>

#include <gea/API.h>

using namespace awds;

bool TcpServer::createSocket() {
	struct sockaddr_in6 addr;
	socklen_t socklen;
	int ret;

	// create an IPv6 socket.
	lsock = socket(PF_INET6, SOCK_STREAM, 0);

	memset(&addr, 0, sizeof(addr));
	addr.sin6_family = AF_INET6;
	addr.sin6_port = htons(this->port);
	memcpy(&addr.sin6_addr, &in6addr_any, sizeof(in6addr_any));
	addr.sin6_flowinfo = 0;

	socklen = sizeof(struct sockaddr_in6);

	/* allow reusing of the same socket */
	int one = 1;
	ret = setsockopt(lsock, SOL_SOCKET,  SO_REUSEADDR, &one, sizeof(int));
	if (ret != 0) {
		close(lsock);
		return false;
	}

	ret = bind(lsock, (struct sockaddr*)&addr, sizeof(struct sockaddr_in6));
	if (ret != 0) {
		close(lsock);
		return false;
	}

	ret = listen(lsock, 4);
	if (ret != 0) {
		close(lsock);
		return false;
	}

	return true;

}

TcpServer::TcpServer(const char *servicename, unsigned short listenport)
		: name(servicename),
		  port(listenport),
		  a_cb(NULL)
{}

bool TcpServer::start() {
	if (!createSocket()) {
		GEA.dbg() << name << ": Cannot open server on port " << port << "!" << std::endl;
		return false;
	}

	lsockHdl = new gea::UnixFdHandle(lsock, gea::PosixModeRead);

}

void TcpServer::waitForClient(AcceptCallback *cb, void *cookie) {
	a_cb = cb;
	a_cookie = cookie;
	GEA.waitFor(lsockHdl, gea::AbsTime::now() + gea::Duration(1000, 1),
	            accept_connection, (void *)this);
}

void TcpServer::accept_connection(gea::Handle *h, gea::AbsTime t, void *data) {
	TcpServer * self = static_cast<TcpServer *>(data);

	assert(self->a_cb);
	if (h->status == gea::Handle::Ready) {

		struct sockaddr_in6 peer_addr;
		socklen_t addr_len = sizeof(struct sockaddr_in6);

		int client_fd = ::accept(self->lsock, (struct sockaddr *)&peer_addr, &addr_len);
		if (client_fd < 0) {
			GEA.dbg() << self->name << ": error in accept: " << strerror(errno) << std::endl;
			return;
		}

		self->a_cb(client_fd, peer_addr, self->a_cookie);
#if 0
		gea::UnixFdHandle *fdHandle = new gea::UnixFdHandle(client_fd, gea::PosixModeRead);
		self->clients[client_fd] = TcpConnection(client_fd, fdHandle, peer);
		GEA.waitFor(fdHandle, t + gea::Duration(1000, 1), read_client_data, data);
#endif
	} else {
		/* reregister accept_connection */
		GEA.waitFor(h, t + gea::Duration(1000, 1), accept_connection, data);
	}
}

TcpConnection::TcpConnection(int fd, struct sockaddr_in6 &peer) :
	fd(fd) {
	char peername[1023];
	char peerport[32];
	socklen_t addr_len = sizeof(struct sockaddr_in6);

	getnameinfo((const struct sockaddr *)&peer, addr_len, peername, sizeof(peername),
			peerport, sizeof(peerport), NI_NUMERICSERV);

	std::ostringstream peerstream;
	peerstream << peername << ":" << peerport;
	std::string peerstring = peerstream.str();
	TcpConnection(fd, peerstring);
	this->peer = peerstring;
	this->fdHdl = new gea::UnixFdHandle(fd, gea::PosixModeRead);
}


void TcpConnection::read_client_data(gea::Handle *h, gea::AbsTime t, void *data) {
	TcpConnection * self = static_cast<TcpConnection *>(data);

	if (h->status == gea::Handle::Ready) {
		char data[2049];
		int ret = h->read(data, sizeof(data)-1);
		if (ret < 0) {
			GEA.dbg() << *self << ".read(): " << strerror(errno) << std::endl;
		}
		if (ret <= 0) {
			if (self->error_cb)
				self->error_cb(*self, NULL, errno, self->cb_cookie);
			else {
				GEA.dbg() << *self << ": short read from client" << std::endl;
			}
			return;
		}
		data[ret] = '\0';
		if (self->recv_cb) {
			self->recv_cb(*self, data, ret, self->cb_cookie);
		} else
			GEA.dbg() << *self << ": " << data << std::endl;
	} else
		GEA.waitFor(h, t + gea::Duration(1000, 1), read_client_data, data);
}

void TcpConnection::waitForData(TcpConnCallback *on_recv, TcpConnCallback *on_error, void *cookie) {
	recv_cb = on_recv;
	error_cb = on_error;
	cb_cookie = cookie;

	GEA.waitFor(fdHdl, gea::AbsTime::now() + gea::Duration(1000, 1), read_client_data, this);
}

void TcpBroadcast::new_client(int client_fd, struct sockaddr_in6 &peer, void *cookie) {
	TcpBroadcast *self = static_cast<TcpBroadcast *>(cookie);

	self->clients[client_fd] = TcpConnection(client_fd, peer);
	GEA.dbg() << self->name << ": new client: " << self->clients[client_fd] << std::endl;

	//GEA.waitFor(fdHandle, t + gea::Duration(1000, 1), read_client_data, data);
	self->clients[client_fd].waitForData(client_recv, client_error, self);
	
	self->waitForClient(new_client, cookie);
}

void TcpBroadcast::client_error(TcpConnection &cl, const char *data, size_t datalen, void *cookie) {
	TcpBroadcast *self = static_cast<TcpBroadcast *>(cookie);

	//GEA.dbg() << self->name << ": closing " << cl << std::endl;
	self->terminate(cl.fd);
}

void TcpBroadcast::client_recv(TcpConnection &cl, const char *data, size_t datalen, void *cookie) {
	TcpBroadcast *self = static_cast<TcpBroadcast *>(cookie);

	GEA.dbg() << self->name << ": " << cl << ": " << data << std::endl;
	cl.waitForData(client_recv, client_error, self);
}

void TcpBroadcast::send_to_all(const char *data, size_t datalen) {
	ClientList::iterator c = clients.begin(), zombie;

	while (c != clients.end()) {
		int ret = ::write(c->first, data, datalen);
		if (ret < 0) {
			GEA.dbg() << this->name << ": write(" << &c->second << "): "
					<< strerror(errno) << std::endl;
			zombie = c;
			c++;
			terminate(zombie->first);
		} else
			c++;
	}
}

void TcpBroadcast::terminate(int fd) {
	TcpConnection &c = clients[fd];
	GEA.dbg() << this->name << ": terminating " << c << std::endl;
	c.close();
	clients.erase(fd);
}

/*
TcpBroadcast::ClientList::iterator TcpBroadcast::findHandle(gea::Handle *h) {
	ClientList::iterator c = clients.begin(), zombie;

	while (c != clients.end()) {
		if (c->second.fdHdl == h) {
			return c;
		} else
			c++;
	}
}

void TcpBroadcast::terminate(int fd) {
	TcpConnection &c = clients[fd];
	GEA.dbg() << this->name << ": terminating " << &c << std::endl;
	c.close();
	clients.erase(fd);
}

void TcpBroadcast::terminate(ClientList::iterator &cli) {
	terminate(cli->first);
}

void TcpBroadcast::terminate(gea::Handle *h) {
	terminate(findHandle(h)->first);
}

void TcpBroadcast::on_receive(RecvCallback *cb, void *cookie) {
	recv_cb = cb;
	recv_cb_cookie = recv_cb_cookie;
}*/
