// vim:ts=4:sw=4:
#ifndef _TCPSERVER_H__
#define _TCPSERVER_H__

#include <map>
#include <iostream>

#include <netinet/in.h>
#include <gea/posix/UnixFdHandle.h>
#include <gea/Time.h>
#include <gea/gea_main.h>

namespace awds {
	class DLLEXPORT TcpConnection {
		std::string peer;
		gea::UnixFdHandle *fdHdl;

		static void read_client_data(gea::Handle *h, gea::AbsTime t, void *data);
	public:
		int fd;

		TcpConnection(int fd, gea::UnixFdHandle *fdHdl, std::string &peer) :
			fd(fd),
			fdHdl(fdHdl),
			peer(peer)
		{}
		TcpConnection(int fd, std::string &peer) {
			TcpConnection(fd, new gea::UnixFdHandle(fd, gea::PosixModeRead), peer);
		}
		TcpConnection(int fd, struct sockaddr_in6 &peer);

		TcpConnection() : fd(-1) {}

		typedef void (TcpConnCallback)(TcpConnection &cl, const char *data, size_t datalen, void *cookie);

		int write(uint8_t *data, ssize_t datalen) {
			return fdHdl->write((char*)data, datalen);
		}
		void close() {
			if (fd >= 0) {
				::close(fd);
				delete fdHdl;
			}
		}

		friend std::ostream& operator<<(std::ostream &s, const TcpConnection &c) {
			return s << c.peer;
		}

		void waitForData(TcpConnCallback *on_recv, TcpConnCallback *on_error, void *cookie);
	private:
		TcpConnCallback *recv_cb, *error_cb;
		void *cb_cookie;
	};

	/** \brief TCP server class for handling incoming TCP connections
	 * \see TopoWatch
	 * \see TcpShell
	 */
	class DLLEXPORT TcpServer {
	protected:
		const char *name;
		unsigned short port;
		int lsock;
		gea::UnixFdHandle *lsockHdl;

		bool createSocket();
		static void accept_connection(gea::Handle *h, gea::AbsTime t, void *data);

	public:
		typedef void (AcceptCallback)(int client_fd, struct sockaddr_in6 &peer, void *cookie);

		TcpServer(const char *servicename, unsigned short listenport);

		virtual bool start();

		void waitForClient(AcceptCallback *cb, void *cookie);
	private:
		AcceptCallback *a_cb;
		void *a_cookie;
	}; // end of class TcpServer

	class DLLEXPORT TcpBroadcast : TcpServer {
		typedef std::map<int, TcpConnection> ClientList;
		ClientList clients;

		ClientList::iterator findHandle(gea::Handle *h);
		//void terminate(ClientList::iterator &cli);
		//void terminate(gea::Handle *h);

		static void new_client(int client_fd, struct sockaddr_in6 &peer, void *cookie);
		static void client_recv(TcpConnection &cl, const char *data, size_t datalen, void *cookie);
		static void client_error(TcpConnection &cl, const char *data, size_t datalen, void *cookie);
	public:
		TcpBroadcast(const char* servicename, unsigned short listenport) :
			TcpServer(servicename, listenport)
		{}

		virtual bool start() {
			if (TcpServer::start()) {
				waitForClient(new_client, this);
				return true;
			} else return false;
		}

		void send_to_all(const char *data, size_t datalen);
		void terminate(int fd);
	};
} // end of namespace awds

#endif // _TCPSERVER_H__

/* This stuff is for emacs
 * Local variables:
 * mode:c++
 * c-basic-offset: 4
 * End:
 */
