#include <cstdlib>
#include <cstdio>

#include <gea/gea_main.h>
#include <gea/ObjRepository.h>
#include <gea/Blocker.h>
#include <gea/API.h>

#include <awds/packettypes.h>
#include <awds/TraceUcPacket.h>
#include <awds/routing.h>
#include <awds/ext/Shell.h>
#include <awds/Topology.h>
#include <awds/toArray.h>

using namespace awds;
using namespace gea;
using namespace std;

/**
 * The pinger module implements the ping protocol. It consists of the server
 * side (ping replier) and the client. When the shell module is found, it adds
 * a 'ping' command.
 */
static const char *short_help = "ping another AWDS node";

static const char *long_help =
    "ping  [options] <DESTINATION>\n"
    "  options: \n"
    "  -t <TTL>   maximum hop count\n"
    "  -s <size>  set packet size\n"
    "  -i <time>  time interval\n"
    "  -p         trace path\n";

/** the Pinger class holds the  state of the pinger module */
struct Pinger {

    bool busy;        ///< flag to indicate that a ping is currently active
    NodeId dest;      ///< the node we are currently pinging
    unsigned ttl;     ///< the maxium number of hops in each ping packet
    int numRepeat;    ///< number of pings to send
    bool doTracePath; ///< flag, if the path should be traced

    // stats.
    double minRTT, maxRTT, sumRTT; ///< measured round trip time 
    int numReceived;    ///< number of received packets
    int numTransmitted; ///< number of transmitted packets

    awds::Routing *awdsRouting; ///< used awds instance
    gea::Blocker blocker;       ///< blocker for creating periodic events
    ostream *out;               ///< ostream to write user messages to
    ShellClient *shellClient;   ///< state of the interactive shell, if present

    Duration period;      ///< perdiod of ping packets
    size_t pingSize;      ///< size of ping packets including header

    AbsTime myT0;         ///< some fized point in time

    Pinger( awds::Routing *awdsRouting):
	awdsRouting(awdsRouting),
	period(2,1)
    {
	busy = false;
	ttl = 44;
	// register ping protocol
	awdsRouting->registerUnicastProtocol(PACKET_TYPE_UC_PING, ping_recv, (void*)this );
	this->out = 0;
	myT0 = gea::AbsTime::now();
    }

    static void next_ping(gea::Handle *h, gea::AbsTime t, void *data);
    static void ping_recv(BasePacket *p,  void *data);

    /** get the ostream for user interaction */
    ostream& dbg() {
	if (out)
	    return *out;
	else
	    return GEA.dbg();
    }

    /** reset the internal state */
    void resetStats();

    void print_stats(std::ostream& out);
    int startPing( Handle *h = 0);
    int parse_opts(int argc, const char* const *argv);
};

/** reset the ping stats.
 */
void Pinger::resetStats() {
    minRTT=5000000.;
    maxRTT=-12;
    sumRTT=0.;
    numReceived = 0;
    numTransmitted = 0;
}

/** parse the given options
 *  \param argc number of arguments
 *  \param argv vector of string pointers
 *  \returns -1 on error, 0 otherwise.
 */
int Pinger::parse_opts(int argc, const char* const *argv) {

    const char *destName = 0;
    char *end_ptr;
    double interval;

    // default settings:
    this->period = Duration(1,1);
    this->pingSize = 200;
    this->ttl=64;
    this->numRepeat = -1;
    this->doTracePath = false;

    for (int i=1; i<argc; ++i) {

	if (!strcmp(argv[i], "-p")) {
	    this->doTracePath = true;
	    continue;
	}

	if (!strcmp(argv[i],"-s")) {
	    ++i;
	    if (i == argc) // no following argument!
		return -1;
	    pingSize = strtol(argv[i], &end_ptr, 0);
	    if (*end_ptr || argv[i] == end_ptr) // invalid conversation
		return -1;
	    continue;
	}

	if (!strcmp(argv[i],"-i")) {
	    ++i;
	    if (i == argc) // no following argument!
		return -1;
	    interval = strtod(argv[i], &end_ptr);
	    if (*end_ptr || argv[i] == end_ptr) // invalid conversation
		return -1;
	    this->period.setSeconds(interval);
	    continue;
	}

	if (!strcmp(argv[i],"-t")) {
	    ++i;
	    if (i == argc) // no following argument!
		return -1;
	    ttl = strtol(argv[i], &end_ptr,0 );
	    if (*end_ptr || argv[i] == end_ptr) // invalid conversation
		return -1;
	    continue;
	}

	// must be the destination name...
	{
	    if (destName) // destination was already given.
		return -1;
	    destName = argv[i];
	}

    } // end for

    if (!destName)
	return -1;

    //    REP_MAP_OBJ(RTopology *, topology);
    dbg() << " dest is " << destName << endl;
    return -!awdsRouting->getNodeByName(this->dest, destName);

} // end parse_opts


/*
   the ping packet format
     ------------------------
    | unicast header         |
     ------------------------
    | trace num              | 2 bytes
     ------------------------
    | ping direction 'i'/'o' | 1 byte
     ------------------------
    | time stamp             | 8 bytes
     ------------------------
    | traced node id 1       | NodeId::size (6) byte
     ------------------------
    | traced node id 2       | NodeId::size (6) byte
     ------------------------
    | ...                    |
*/

/* the following constansts define fixed offsets in the ping packet */
static const size_t PingTraceNum  = UnicastPacket::UnicastPacketEnd;
static const size_t PingDirection = PingTraceNum + 2;
static const size_t PingTimeStamp = PingDirection + 1;
static const size_t PingHeaderEnd = PingTimeStamp + 8;

/** send the next ping packet
 */
void Pinger::next_ping(gea::Handle *h, gea::AbsTime t, void *data) {

    Pinger *self = static_cast<Pinger *>(data);

    if (h->status == gea::Handle::Ready) {
	// key pressed in interactive mode!
	self->shellClient->unblock();
	self->busy = false;
	self->print_stats( self->dbg() );
	return;
    }

    BasePacket *p = self->awdsRouting->newUnicastPacket(PACKET_TYPE_UC_PING);
    p->setDest(self->dest);

    TraceUcPacket uniP(*p);

    uniP.setUcDest(self->dest);
    uniP.setTTL(self->ttl);
    self->pingSize = max(self->pingSize,  PingHeaderEnd);

    p->size = self->pingSize;
    p->buffer[PingDirection] = 'i';
    (t - self->myT0).toArray( &p->buffer[PingTimeStamp] );

    if (self->doTracePath) {
	uniP.setTracePointer(PingHeaderEnd);
	uniP.setTraceFlag(1);
    }

    self->dbg() << "sending ping to " << self->dest
	       << std::endl;

    self->awdsRouting->sendUnicast(p);
    p->unref();

    self->numTransmitted++;

    GEA.waitFor( h,
		 t + self->period,
		 next_ping,
		 data);

}

/** prints summary after stopping the ping session
 *
 *  \param out ostream to write the text to.
 */
void Pinger::print_stats(std::ostream& out) {
    out << "--- " << dest << " ping statistics ---" << endl
	<< numTransmitted << "  packets transmitted, " << numReceived << " received, "
	<< ( 100. * double(numTransmitted - numReceived) / numTransmitted) << "% packet loss" << endl
	<< "rtt min/avg/max = " << minRTT << "/" << (sumRTT / numReceived) << "/" << maxRTT
	<< endl;
}

/** called when a ping packet is received */
void Pinger::ping_recv(BasePacket *p, void *data) {

    Pinger *self = static_cast<Pinger *>(data);

    TraceUcPacket uniP(*p);

    if ( p->buffer[PingDirection] == 'i' ) {

	GEA.dbg() << "received ping from " << uniP.getSrc()
		  << " ttl=" << uniP.getTTL()
		  << std::endl;

	uniP.setUcDest(uniP.getSrc());
	uniP.setSrc(self->awdsRouting->myNodeId);
	p->buffer[PingDirection] = 'o';
	p->ref();
	self->awdsRouting->sendUnicast(p);
	p->unref();
    } else /* must be pong */ {

	const NodeId src = uniP.getSrc();

	Duration timestamp;
	timestamp.fromArray(&p->buffer[PingTimeStamp]); // read timestamp from packet

	double deltaT= (GEA.lastEventTime - ( self->myT0 + timestamp ) ).getNanoSecsD();
	ostream& out = self->dbg();

	out << "received pong from " << src
	    << " dur=" << (deltaT/1000.) << " usec."
	    << " ttl=" << uniP.getTTL();

	if (uniP.getTraceFlag()) {

	    const size_t tPtr = uniP.getTracePointer();

	    out << "\tpath[" << ( (tPtr - PingHeaderEnd) / NodeId::size ) <<"]";
	    for ( size_t nPtr = PingHeaderEnd;
		  nPtr < tPtr;
		  nPtr += NodeId::size) {

		NodeId node;
		node.fromArray(p->buffer + nPtr);
		out << " " << node;
	    }
	}
	out << endl;

	// collect stats
	self->minRTT = min(self->minRTT, deltaT);
	self->maxRTT = max(self->maxRTT, deltaT);
	self->sumRTT += deltaT;
	self->numReceived++;

    }

}

/** The shell command hook of the 'ping' command. it is called when 
 *  'ping ...' is entered into the AWDS command shell.
 *  
 *  \param sc represents the connected shell.
 *  \param data the object pointer.
 *  \param argc number of arguments given to the ping command.
 *  \param argv arguments given to the ping command.
 */
static int ping_command_fn(ShellClient &sc, void *data, int argc, char **argv) {
    Pinger *self = static_cast<Pinger*>(data);

    if (argc <= 1) { // missing arguments
	*sc.sockout << long_help << endl;
	return -1;
    }

    if (self->busy) { // a ping is already active on another shell
	*sc.sockout << "ping already running" << endl;
	return 0;
    }

    int ret = self->parse_opts(argc,argv); // parse the arguments
    if (ret != 0) { // error on parsing the arguments
	*sc.sockout << long_help << endl;
	return -1;
    }

    if ( self->busy ) { // check twice?
	*sc.sockout << "pinger is already in use " << endl;
	return -1;
    }

    self->shellClient = &sc; // store the client connection for later use
    self->out = sc.sockout;  // store the output stream for later use
    sc.block();              // block the shell, s.th. we get control of the input.

    ret = self->startPing(sc.sockin); // start the pinging.

    return 0; // report success
}

/** add the 'ping' command to the shell. */
void add_shell(Pinger *pinger) {
    REP_MAP_OBJ(Shell *, shell);

    if (!shell) return;

    shell->add_command("ping", ping_command_fn, pinger, short_help, long_help);

}

/** start ping
 *
 *  start the client side of the ping protocol
 *  \param h handle of interactive shell, used for generating periodic events
 *           and for stopping the ping
 *  \returns -1 on error, 0 on success
 */
int Pinger::startPing( Handle *h) {

    if (busy) // never start when busy
	return -1;

    busy = true; // we are busy now, prevents start of  second ping
    if (h == 0)  // ping is started without interactive shell
	h = &this->blocker; // use blocker in this case

    resetStats(); // reset ping stats
    dbg() << "pinging " << dest << endl; // print some meaningfull message

    // start the  ping events
    GEA.waitFor(h,
		GEA.lastEventTime + gea::Duration(1,10000) /* start in 100 us */,
		next_ping, static_cast<void *>(this));
    return 0;
}

/** the entry point of the pinger module
 */
GEA_MAIN_2(pinger, argc, argv)
{
    for (int i=0;i<argc;++i) {
	string w(argv[i]);
	
	if (w == "--help") {
	    return -1;
	}
    }

    Pinger *pinger; // the pinger state
    REP_MAP_OBJ(awds::Routing *, routing); // get routing object from object repository

    if (!routing) {
	GEA.dbg() << "cannot find object 'routing' in repository" << std::endl;
	return -1;
    }

    pinger = new Pinger(routing);

    add_shell(pinger);

    if (argc > 1) {
	pinger->parse_opts(argc, argv);
	pinger->startPing();
    }

    return 0;
}
/* This stuff is for emacs
 * Local variables:
 * mode:c++
 * c-basic-offset: 4
 * End:
 */
